package com.github.f4b6a3.ulid.creator;

import java.util.Random;
import java.util.UUID;

import org.junit.Test;

import com.github.f4b6a3.commons.random.Xorshift128PlusRandom;
import com.github.f4b6a3.ulid.exception.UlidCreatorException;
import com.github.f4b6a3.ulid.timestamp.FixedTimestampStretegy;

import static org.junit.Assert.*;

public class UlidBasedGuidCreatorTest {

	private static final long DEFAULT_LOOP_MAX = 1_000_000;

	private static final long TIMESTAMP = System.currentTimeMillis();

	private static final Random RANDOM = new Xorshift128PlusRandom();

	@Test
	public void testRandomMostSignificantBits() {

		UlidBasedGuidCreatorMock creator = new UlidBasedGuidCreatorMock(TIMESTAMP);
		creator.withTimestampStrategy(new FixedTimestampStretegy(TIMESTAMP));

		UUID uuid = creator.createGuid();
		long firstMsb = creator.extractRandomMsb(uuid);
		for (int i = 0; i < DEFAULT_LOOP_MAX; i++) {
			uuid = creator.createGuid();

		}

		long lastMsb = creator.extractRandomMsb(uuid);
		long expectedMsb = firstMsb;
		assertEquals(String.format("The last MSB should be iqual to the first %s.", expectedMsb), expectedMsb, lastMsb);

		creator.withTimestampStrategy(new FixedTimestampStretegy(TIMESTAMP + 1));
		uuid = creator.createGuid();
		lastMsb = uuid.getMostSignificantBits();
		assertNotEquals("The last MSB should be random after timestamp changed.", firstMsb, lastMsb);
	}

	@Test
	public void testRandomLeastSignificantBits() {

		UlidBasedGuidCreatorMock creator = new UlidBasedGuidCreatorMock(TIMESTAMP);
		creator.withTimestampStrategy(new FixedTimestampStretegy(TIMESTAMP));

		UUID uuid = creator.createGuid();
		long firstLsb = creator.extractRandomLsb(uuid);
		for (int i = 0; i < DEFAULT_LOOP_MAX; i++) {
			uuid = creator.createGuid();
		}

		long lastLsb = creator.extractRandomLsb(uuid);
		long expected = firstLsb + DEFAULT_LOOP_MAX;
		assertEquals(String.format("The last LSB should be iqual to %s.", expected), expected, lastLsb);

		long notExpected = firstLsb + DEFAULT_LOOP_MAX + 1;
		creator.withTimestampStrategy(new FixedTimestampStretegy(TIMESTAMP + 1));
		uuid = creator.createGuid();
		lastLsb = uuid.getLeastSignificantBits();
		assertNotEquals("The last LSB should be random after timestamp changed.", notExpected, lastLsb);
	}

	@Test
	public void testIncrementOfRandomLeastSignificantBits() {

		UlidBasedGuidCreatorMock creator = new UlidBasedGuidCreatorMock(TIMESTAMP);
		creator.withTimestampStrategy(new FixedTimestampStretegy(TIMESTAMP));

		long lsb = creator.getRandomLsb();

		UUID uuid = new UUID(0, 0);
		for (int i = 0; i < DEFAULT_LOOP_MAX; i++) {
			uuid = creator.createGuid();
		}

		long expectedLsb = lsb + DEFAULT_LOOP_MAX;
		long randomLsb = creator.getRandomLsb();
		assertEquals("Wrong LSB after loop.", expectedLsb, randomLsb);

		randomLsb = creator.extractRandomLsb(uuid);
		assertEquals("Wrong LSB after loop.", expectedLsb, randomLsb);
	}

	@Test
	public void testIncrementOfRandomMostSignificantBits() {

		UlidBasedGuidCreatorMock creator = new UlidBasedGuidCreatorMock(TIMESTAMP);
		creator.withTimestampStrategy(new FixedTimestampStretegy(TIMESTAMP));

		long msb = creator.getRandomMsb();

		UUID uuid = new UUID(0, 0);
		for (int i = 0; i < DEFAULT_LOOP_MAX; i++) {
			uuid = creator.createGuid();
		}

		long expectedMsb = msb;
		long randomMsb = creator.getRandomMsb();
		assertEquals("Wrong MSB after loop.", expectedMsb, randomMsb);

		randomMsb = creator.extractRandomMsb(uuid);
		assertEquals("Wrong MSB after loop.", expectedMsb, randomMsb);
	}

	@Test
	public void testShouldThrowOverflowException1() {

		long msbMax = 0x000001ffffffffffL;
		long lsbMax = 0x000001ffffffffffL;

		long msb = msbMax - 1;
		long lsb = lsbMax - DEFAULT_LOOP_MAX;

		UlidBasedGuidCreatorMock creator = new UlidBasedGuidCreatorMock(msb, lsb, msbMax, lsbMax, TIMESTAMP);
		creator.withTimestampStrategy(new FixedTimestampStretegy(TIMESTAMP));

		for (int i = 0; i < DEFAULT_LOOP_MAX - 1; i++) {
			creator.createGuid();
		}

		try {
			creator.createGuid();
			fail("It should throw an overflow exception.");
		} catch (UlidCreatorException e) {
			// success
		}
	}

	@Test
	public void testShouldThrowOverflowException2() {

		long msbMax = (RANDOM.nextLong() & UlidBasedGuidCreatorMock.HALF_RANDOM_COMPONENT)
				| UlidBasedGuidCreatorMock.INCREMENT_MAX;
		long lsbMax = (RANDOM.nextLong() & UlidBasedGuidCreatorMock.HALF_RANDOM_COMPONENT)
				| UlidBasedGuidCreatorMock.INCREMENT_MAX;

		long msb = msbMax - 1;
		long lsb = lsbMax - DEFAULT_LOOP_MAX;

		UlidBasedGuidCreatorMock creator = new UlidBasedGuidCreatorMock(msb, lsb, msbMax, lsbMax, TIMESTAMP);
		creator.withTimestampStrategy(new FixedTimestampStretegy(TIMESTAMP));

		UUID uuid = new UUID(0, 0);
		for (int i = 0; i < DEFAULT_LOOP_MAX - 1; i++) {
			uuid = creator.createGuid();
		}

		long expectedLsb = (lsbMax - 1) & UlidBasedGuidCreatorMock.HALF_RANDOM_COMPONENT;
		long randomLsb = creator.extractRandomLsb(uuid);
		assertEquals("Incorrect LSB after loop.", expectedLsb, randomLsb);

		long expectedMsb = (msbMax - 1) & UlidBasedGuidCreatorMock.HALF_RANDOM_COMPONENT;
		long randomMsb = creator.extractRandomMsb(uuid);
		assertEquals("Incorrect MSB after loop.", expectedMsb, randomMsb);

		try {
			creator.createGuid();
			fail("It should throw an overflow exception.");
		} catch (UlidCreatorException e) {
			// success
		}
	}
}
